# sample use: 
# python inclass-robot-final.py 16 5 7 0 12 4 "2,2;1,1"
# python inclass-robot-final.py 5 5 0 0 4 0 "2,0;2,1;2,2"
# see the section below on parsing obstacles for more details

from z3 import *
import sys

# --------------------------------------------------------
# Parse command line args
# Note that several aren't used in this program but will be required for HW.
# And one is required for this program but should be removed for HW.
# (See more below!)
# --------------------------------------------------------

def parse_obstacles(str):
	coords = str.split(";")
	return [(int(comps[0]), int(comps[1])) for comps in [coord.split(",") for coord in coords]]

grid_width = int(sys.argv[1]) # how many squares across?
grid_length = int(sys.argv[2]) # how many squares up and down?  (ignored in this prog, but required for HW)
start_x_coord = int(sys.argv[3]) # with leftmost colum at coord 0, coord of start square
start_y_coord = int(sys.argv[4]) # with topmost colum at coord 0, coord of start square  (ignored in this prog, but required for HW)
goal_x_coord = int(sys.argv[5]) # with leftmost colum at coord 0, coord of goal square
goal_y_coord = int(sys.argv[6]) # with topmost colum at coord 0, coord of goal square  (ignored in this prog, but required for HW)
obstacles = parse_obstacles(sys.argv[7]) #(ignored in this prog, but required for HW)
num_instrs = int(sys.argv[8]) # how many instructions will we allow in our output program?  DO NOT USE IN HW!!!  (but required for this prog)

# --------------------------------------------------------
# The fun stuff :)
# --------------------------------------------------------

# what's the effect of running one instruction?  where does the robot end up?
def run_instr(pos, instr, arg, envir):
	return If(arg > 0,
			If(instr == 0,
				# left
				If(pos - arg >= 0, pos - arg, 0),
			If(instr == 1,
				# right
				If(pos + arg <= (envir - 1), pos + arg, envir - 1),
			pos)),
			pos)

# what's the effect of running the whole program?  where does the robot end up?
def run_prog(pos, instrs, args, envir):
	curr_pos = pos
	for i in range(len(instrs)):
		curr_pos = run_instr(curr_pos, instrs[i], args[i], envir)
	return curr_pos

# let's make some Z3 bitvectors that we'll use to search the space of instructions
def gen_instrs(num_instrs):
	return [BitVec('x_'+str(i),2) for i in range(num_instrs)]

# let's make some Z3 bitvectors that we'll use to search the space of arguments
def gen_args(num_instrs):
	return [Int('a_'+str(i)) for i in range(num_instrs)]

# a convenience function for printing the Z3 output to look like a sequence of instructions
def print_model(model, instrs, args):
	for i in range(len(instrs)):
		val = model.eval(instrs[i])
		arg = model.eval(args[i])
		if (val == 0):
			print("L", arg)
		elif (val == 1):
			print("R", arg)
		else:
			print("-")


instrs = gen_instrs(num_instrs) # generate BVs to represent instructions
args = gen_args(num_instrs) # generate BVs to represent arguments
goal = run_prog(start_x_coord, instrs, args, grid_width) == goal_x_coord # where do we want our robot to move?

s = Solver()
s.add(goal)
satisfiable = s.check()
# print("satisfiable?", satisfiable)

if (satisfiable == sat):
	model = s.model()
	print_model(model, instrs, args) # print the program if we found one
	# print(model) # print the model, just to visualize what's happening underneath
